@@ -7,6 +7,7 @@
 Classes:
 
 Model                  - Modeling with any function using Levenberg-Marquardt.
+Parameter              - Information about a specific model parameter.
 PolynomialModel        - Modeling with polynomials.
 ScaleModel             - Modeling with a single scale factor.
 ComposedModel          - Modeling with combinations of pluggable components.
@@ -51,20 +52,68 @@ try:
 except ImportError:
     import numpy.polynomial as npoly
 
+from . import binary_type, text_type
+
+
+class Parameter (object):
+    """Information about a parameter in a least-squares model. These data may only
+    be obtained after solving least-squares problem.
+
+    These objects reference information from their parent objects, so changing
+    the parent will alter the apparent contents of these objects.
+
+    Properties:
+
+    index  - The parameter's index in the Model's arrays.
+    name   - The parameter's name.
+    value  - The parameter's value.
+    uncert - The uncertainty in value.
+    uval   - Accesses ``value`` and ``uncert`` as a ``pwkit.msmt.Uval``.
+
+    """
+    def __init__ (self, owner, index):
+        self._owner = owner
+        self._index = index
+
+    def __repr__ (self):
+        return '<Parameter "%s" (#%d) of %s>' % (self.name, self._index, self._owner)
+
+    @property
+    def index (self): # make this read-only
+        return self._index
+
+    @property
+    def name (self):
+        return self._owner.pnames[self._index]
+
+    @property
+    def value (self):
+        return self._owner.params[self._index]
+
+    @property
+    def uncert (self):
+        return self._owner.puncerts[self._index]
+
+    @property
+    def uval (self):
+        from .msmt import Uval
+        return Uval.from_norm (self.value, self.uncert)
+
 
 class _ModelBase (object):
+
     """Data and a model for least-squares fitting. Attributes:
 
-    data       - The data to be modeled.
-    invsigma   - Weights: 1/σ for each data point.
-    params     - ndarray of solved model parameters.
-    perror     - ndarray of 1-sigma uncerts on params.
-    paramnames - iterable of string names for parameters.
-    covar      - Covariance matrix of parameters.
-    mfunc      - A function f(...) evaluating model fixed at best params.
-    mdata      - The modeled data at the best params.
-    rchisq     - Reduced χ² of the fit.
-    resids     - resids = data - mdata.
+    data     - The data to be modeled.
+    invsigma - Weights: 1/σ for each data point.
+    params   - ndarray of solved model parameters.
+    puncerts - ndarray of 1-sigma uncerts on params.
+    pnames   - list of string names for parameters.
+    covar    - Covariance matrix of parameters.
+    mfunc    - A function f(...) evaluating model fixed at best params.
+    mdata    - The modeled data at the best params.
+    rchisq   - Reduced χ² of the fit.
+    resids   - resids = data - mdata.
 
     Methods:
 
@@ -75,17 +124,24 @@ class _ModelBase (object):
     show_corr  - Show the parameter correlation matrix with `pwkit.ndshow_gtk2`.
     solve      - Fit the model to the data.
 
+    A ``Parameter`` data structure may be obtained by indexing this object
+    with either the parameter's numerical index or its name. I.e.,
+
+      m = Model (...).solve (...)
+      p = m['slope']
+      print (p.name, p.value, p.uncert, p.uval)
+
     """
     data = None
     invsigma = None
-    params = None # ndarray of solved model parameters
-    perror = None # ndarray of 1-sigma uncerts on pamams
-    paramnames = None # iterable of string names for parameters
-    covar = None # covariance matrix
-    mfunc = None # function f(...) evaluating model fixed at best params
-    mdata = None # modeled data at best params
-    rchisq = None # reduced chi squared of the fit
-    resids = None # resids = data - mdata
+    params = None
+    puncerts = None
+    pnames = None
+    covar = None
+    mfunc = None
+    mdata = None
+    rchisq = None
+    resids = None
 
     def __init__ (self, data, invsigma=None):
         self.set_data (data, invsigma)
@@ -105,13 +161,13 @@ class _ModelBase (object):
 
 
     def print_soln (self):
-        lmax = reduce (max, (len (x) for x in self.paramnames), len ('r chi sq'))
+        lmax = reduce (max, (len (x) for x in self.pnames), len ('r chi sq'))
 
-        if self.perror is None:
-            for pn, val in zip (self.paramnames, self.params):
+        if self.puncerts is None:
+            for pn, val in zip (self.pnames, self.params):
                 print ('%s: %14g' % (pn.rjust (lmax), val))
         else:
-            for pn, val, err in zip (self.paramnames, self.params, self.perror):
+            for pn, val, err in zip (self.pnames, self.params, self.puncerts):
                 frac = abs (100. * err / val)
                 print ('%s: %14g +/- %14g (%.2f%%)' % (pn.rjust (lmax), val, err, frac))
 
@@ -122,6 +178,27 @@ class _ModelBase (object):
         return self
 
 
+    def __getitem__ (self, key):
+        if isinstance (key, binary_type):
+            # If you're not using the unicode_literals __future__, things get
+            # annoying really quickly without this.
+            key = text_type (key)
+
+        if isinstance (key, int):
+            idx = key
+            if idx < 0 or idx >= len (self.pnames):
+                raise ValueError ('illegal parameter number %d' % key)
+        elif isinstance (key, text_type):
+            try:
+                idx = self.pnames.index (key)
+            except ValueError:
+                raise ValueError ('no such parameter named "%s"' % key)
+        else:
+            raise ValueError ('illegal parameter key %r' % key)
+
+        return Parameter (self, idx)
+
+
     def plot (self, modelx, dlines=False, xmin=None, xmax=None,
               ymin=None, ymax=None, **kwargs):
         """This assumes that `data` is 1D and that `mfunc` takes one argument that
@@ -181,7 +258,7 @@ class Model (_ModelBase):
             self.set_data (data, invsigma)
 
 
-    def set_func (self, func, npar, parnames, args=()):
+    def set_func (self, func, pnames, args=()):
         """This function creates the `lmmin.Problem` so that the caller can futz with
         it before calling solve(), if so desired.
 
@@ -190,19 +267,33 @@ class Model (_ModelBase):
 
         self.func = func
         self._args = args
-        self.lm_prob = Problem (npar)
-        self.paramnames = parnames
+        self.pnames = list (pnames)
+        self.lm_prob = Problem (len (self.pnames))
         return self
 
 
     def set_simple_func (self, func, args=()):
         npar = func.func_code.co_argcount - len (args)
-        parnames = func.func_code.co_varnames[:npar]
+        pnames = func.func_code.co_varnames[:npar]
 
         def wrapper (params, *args):
             return func (*(tuple (params) + args))
 
-        return self.set_func (wrapper, npar, parnames, args)
+        return self.set_func (wrapper, pnames, args)
+
+
+    def make_frozen_func (self, params):
+        """Returns a model function frozen to the specified parameter values. Any
+        remaining arguments are left free and must be provided when the
+        function is called.
+
+        This just applies `functools.partial` to the `func` property of this
+        object.
+
+        """
+        params = np.array (params, dtype=np.float, ndmin=1)
+        from functools import partial
+        return partial (self.func, params)
 
 
     def solve (self, guess):
@@ -219,11 +310,9 @@ class Model (_ModelBase):
         self.lm_soln = soln = self.lm_prob.solve (guess)
 
         self.params = soln.params
-        self.perror = soln.perror
+        self.puncerts = soln.perror
         self.covar = soln.covar
-
-        from functools import partial
-        self.mfunc = partial (f, *soln.params)
+        self.mfunc = self.make_frozen_func (soln.params)
 
         # fvec = resids * invsigma = (data - mdata) * invsigma
         self.resids = soln.fvec.reshape (self.data.shape) / self.invsigma
@@ -232,10 +321,30 @@ class Model (_ModelBase):
         if soln.ndof > 0:
             # lm_soln.fnorm can be unreliable ("max (fnorm, fnorm1)" branch)
             self.rchisq = (self.lm_soln.fvec**2).sum () / soln.ndof
+        else:
+            self.rchisq = None
         return self
 
 
 class PolynomialModel (_ModelBase):
+    """Least-squares polynomial fit.
+
+    Because this is a very specialized kind of problem, we don't need an
+    initial guess to solve, and we can use fast built-in numerical routines.
+
+    The output parameters are named "a0", "a1", ... and are stored in that
+    order in PolynomialModel.params[]. We have ``y = sum(x**i * a[i])``, so
+    "a2" = "params[2]" is the quadratic term, etc.
+
+    This model does *not* give uncertainties on the derived coefficients. The
+    as_nonlinear() method can be use to get a `Model` instance with
+    uncertainties.
+
+    Methods:
+
+    as_nonlinear - Return a (lmmin-based) `Model` equivalent to self.
+
+    """
     def __init__ (self, maxexponent, x, data, invsigma=None):
         self.maxexponent = maxexponent
         self.x = np.array (x, dtype=np.float, ndmin=1, copy=False, subok=True)
@@ -243,21 +352,51 @@ class PolynomialModel (_ModelBase):
 
 
     def solve (self):
-        self.paramnames = ['a%d' % i for i in xrange (self.maxexponent + 1)]
+        self.pnames = ['a%d' % i for i in xrange (self.maxexponent + 1)]
         # Based on my reading of the polyfit() docs, I think w=invsigma**2 is right...
         self.params = npoly.polyfit (self.x, self.data, self.maxexponent,
                                      w=self.invsigma**2)
-        self.perror = None # does anything provide this? could farm out to lmmin ...
+        self.puncerts = None # does anything provide this? could farm out to lmmin ...
         self.covar = None
         self.mfunc = lambda x: npoly.polyval (x, self.params)
         self.mdata = self.mfunc (self.x)
         self.resids = self.data - self.mdata
-        self.rchisq = (((self.resids * self.invsigma)**2).sum ()
-                       / (self.x.size - (self.maxexponent + 1)))
+
+        if self.x.size > self.maxexponent + 1:
+            self.rchisq = (((self.resids * self.invsigma)**2).sum ()
+                           / (self.x.size - (self.maxexponent + 1)))
+        else:
+            self.rchisq = None
+
         return self
 
 
+    def as_nonlinear (self, params=None):
+        """Return a `Model` equivalent to this object. The nonlinear solver is less
+        efficient, but lets you freeze parameters, compute uncertainties, etc.
+
+        If the `params` argument is provided, solve() will be called on the
+        returned object with those parameters. If it is `None` and this object
+        has parameters in `self.params`, those will be use. Otherwise, solve()
+        will not be called on the returned object.
+
+        """
+        if params is None:
+            params = self.params
+
+        nlm = Model (None, self.data, self.invsigma)
+        nlm.set_func (lambda p, x: npoly.polyval (x, p),
+                      self.pnames,
+                      args=(self.x,))
+
+        if params is not None:
+            nlm.solve (params)
+        return nlm
+
+
 class ScaleModel (_ModelBase):
+    """Solve `data = m * x` for `m`."""
+
     def __init__ (self, x, data, invsigma=None):
         self.x = np.array (x, dtype=np.float, ndmin=1, copy=False, subok=True)
         self.set_data (data, invsigma)
@@ -269,10 +408,10 @@ class ScaleModel (_ModelBase):
         m = sxy / sxx
         uc_m = 1. / np.sqrt (sxx)
 
-        self.paramnames = ['m']
+        self.pnames = ['m']
         self.params = np.asarray ([m])
-        self.perror = np.asarray ([uc_m])
-        self.covar = self.perror.reshape ((1, 1))
+        self.puncerts = np.asarray ([uc_m])
+        self.covar = self.puncerts.reshape ((1, 1))
         self.mfunc = lambda x: m * x
         self.mdata = m * self.x
         self.resids = self.data - self.mdata
@@ -286,7 +425,7 @@ class ScaleModel (_ModelBase):
 class ModelComponent (object):
     npar = 0
     name = None
-    paramnames = ()
+    pnames = ()
     nmodelargs = 0
 
     setguess = None
@@ -300,7 +439,7 @@ class ModelComponent (object):
     def _param_names (self):
         """Overridable in case the list of parameter names needs to be
         generated on the fly."""
-        return self.paramnames
+        return self.pnames
 
     def finalize_setup (self):
         """If the component has subcomponents, this should set their `name`,
@@ -388,7 +527,7 @@ class ComposedModel (_ModelBase):
         self.lm_prob = Problem (component.npar)
         self.force_guess = np.empty (component.npar)
         self.force_guess.fill (np.nan)
-        self.paramnames = list (component._param_names ())
+        self.pnames = list (component._param_names ())
 
         component.prep_params ()
 
@@ -414,7 +553,7 @@ class ComposedModel (_ModelBase):
         self.lm_soln = soln = self.lm_prob.solve (guess)
 
         self.params = soln.params
-        self.perror = soln.perror
+        self.puncerts = soln.perror
         self.covar = soln.covar
 
         # fvec = resids * invsigma = (data - mdata) * invsigma
@@ -446,7 +585,7 @@ class ComposedModel (_ModelBase):
 
 class AddConstantComponent (ModelComponent):
     npar = 1
-    paramnames = ('value', )
+    pnames = ('value', )
     nmodelargs = 0
 
     def model (self, pars, mdata):
@@ -473,6 +612,7 @@ class AddValuesComponent (ModelComponent):
     nmodelargs = 0
 
     def __init__ (self, nvals, name=None):
+        super (AddValuesComponent, self).__init__ (name)
         self.npar = nvals
 
     def _param_names (self):
@@ -502,6 +642,7 @@ class AddPolynomialComponent (ModelComponent):
     nmodelargs = 1
 
     def __init__ (self, maxexponent, x, name=None):
+        super (AddPolynomialComponent, self).__init__ (name)
         self.npar = maxexponent + 1
         self.x = np.array (x, dtype=np.float, ndmin=1, copy=False, subok=True)
 
@@ -569,7 +710,7 @@ class SeriesComponent (ModelComponent):
     is only valid if every subcomponent except the first is additive --
     otherwise, the Jacobian won't be right."""
 
-    def __init__ (self, name=None, components=()):
+    def __init__ (self, components=(), name=None):
         super (SeriesComponent, self).__init__ (name)
         self.components = list (components)
 
@@ -865,7 +1006,7 @@ class MatMultComponent (ModelComponent):
 class ScaleComponent (ModelComponent):
     npar = 1
 
-    def __init__ (self, name=None, subcomp=None):
+    def __init__ (self, subcomp=None, name=None):
         super (ScaleComponent, self).__init__ (name)
         self.setsubcomp (subcomp)
 
